
'''
This file is automatically generated; Do not edit it. :)
'''
VERSION_INFO = {
    'final': True,
    'version': '1.22.4',
    'branch_nick': 'patch-0-Release__1_22_4_preparation',
    'revision_id': 'e6a39c8edde815a3b88d6f0f4d2f1cceb867ec08',
    'revno': 9451
}
